#==========================================================
# Functions --
#
#   procedures for manipulating PostgreSQL Functions
#
#==========================================================
#
namespace eval Functions {
    variable Win
    variable name
    variable nametodrop
    variable parameters
    variable parameterstodrop
    variable returns
    variable returnstodrop
    variable language
    variable comment
}


#----------------------------------------------------------
# ::Functions::new --
#
#   Prepare the function window for a new function.
#----------------------------------------------------------
#
proc ::Functions::new {} {

    variable Win
    variable name
    variable nametodrop
    variable parameters
    variable parameterstodrop
    variable returns
    variable returnstodrop
    variable language
    variable comment

    set name ""
    set nametodrop ""
    set parameters ""
    set parameterstodrop ""
    set returns ""
    set returnstodrop ""
    set language ""

    Window show .pgaw:Function

    $Win(fnctxt) delete 1.0 end
    $Win(comment) delete 1.0 end
    focus $Win(name)

}; # end proc ::Functions::new


#----------------------------------------------------------
# ::Functions::open --
#
#   Open the function in design edit mode.
#
# Parameters:
#   functionname_   name of the function w/parameter types
#----------------------------------------------------------
#
proc ::Functions::open {functionname_} {

    ::Functions::design $functionname_

}; # end proc ::Functions::open


#----------------------------------------------------------
# ::Functions::design --
#
#   Open the function in design edit mode.  Does the work of
#   obtaining the SQL definition of the function.
#
# Parameters:
#   functionname_   name of the function w/parameter types
#----------------------------------------------------------
#
proc ::Functions::design {functionname_} {

    global CurrentDB
    variable Win
    variable name
    variable nametodrop
    variable returns
    variable returnstodrop
    variable parameters
    variable parameterstodrop
    variable language
    variable comment

    # if we are schema qualified, strip out the schema name
    set schema [string range $functionname_ 0 [expr {[string first . $functionname_]-1}]]
    if { [string length $schema] > 0 } {
        set functionname_ [string range $functionname_ [expr {[string length $schema]+1}] end]
    }

    Window show .pgaw:Function
    $Win(fnctxt) delete 1.0 end

    if { [string length $schema] > 0 } {
        set sql "SELECT p.*,
                        n.nspname || '.' || p.proname
                     AS proname,
                        obj_description(p.oid)
                     AS fdesc
                   FROM pg_catalog.pg_proc p,
                        pg_catalog.pg_namespace n
                  WHERE
                        p.pronamespace=n.oid
                    AND
                    (SELECT SUBSTRING('$functionname_'
                       FROM 1
                        FOR POSITION('(' IN '$functionname_')-1)=proname)
                    AND
                    (SELECT SUBSTRING('$functionname_'
                       FROM POSITION('(' IN '$functionname_')+1
                            FOR LENGTH('$functionname_')-POSITION('(' IN '$functionname_')-1)=OIDVECTORTYPES(proargtypes))"
    } else {
        set sql "SELECT p.*,
                        obj_description(p.oid)
                     AS fdesc
                   FROM [::Database::qualifySysTable pg_proc] p
                  WHERE
                    (SELECT SUBSTRING('$functionname_'
                       FROM 1
                        FOR POSITION('(' IN '$functionname_')-1)=proname)
                    AND
                    (SELECT SUBSTRING('$functionname_'
                       FROM POSITION('(' IN '$functionname_')+1
                            FOR LENGTH('$functionname_')-POSITION('(' IN '$functionname_')-1)=OIDVECTORTYPES(proargtypes))"
    }

    wpg_select $CurrentDB $sql rec {
        set name $rec(proname)
        set temppar $rec(proargtypes)
        set returns [::Database::getPgType $rec(prorettype)]
        if {$returns == "unknown"} {
            set returns "opaque"
        }
        set returnstodrop $returns
        set funcnrp $rec(pronargs)
        set prolanguage $rec(prolang)
        set comment $rec(fdesc)
        $Win(comment) insert end $rec(fdesc)
        $Win(fnctxt) insert end $rec(prosrc)
    }

    set sql "
        SELECT lanname
          FROM [::Database::qualifySysTable pg_language]
         WHERE oid=$prolanguage"

    wpg_select $CurrentDB "$sql" rec {
        set language $rec(lanname)
    }
    if { $language=="C" || $language=="c" } {

    set sql "
        SELECT probin
          FROM [::Database::qualifySysTable pg_proc]
         WHERE proname='$functionname_'"

        wpg_select $CurrentDB "$sql" rec {
            $Win(fnctxt) delete 1.0 end
            $Win(fnctxt) insert end $rec(probin)
        }
    }
    set lparams [list]
    for {set i 0} {$i<$funcnrp} {incr i} {
        lappend lparams [Database::getPgType [lindex $temppar $i]]
    }
    set parameters [join $lparams ,]
    set nametodrop $name
    set parameterstodrop $parameters

    Syntax::highlight $Win(fnctxt) $language

}; # end proc ::Functions::design


#----------------------------------------------------------
# ::Functions::save --
#
#   saves the function
#
# Arguments:
#   saveas_     1 if this should be a "save as" instead
#               of the default save
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Functions::save {{saveas_ 0}} {

    variable Win
    variable name
    variable nametodrop
    variable returns
    variable returnstodrop
    variable parameters
    variable parameterstodrop
    variable language

    if {$name==""} {
        focus $Win(name)
        showError [intlmsg "You must supply a name for this function!"]
    } elseif {$returns==""} {
        focus $Win(returns)
        showError [intlmsg "You must supply a return type!"]
    } elseif {$language==""} {
        focus $Win(language)
        showError [intlmsg "You must supply the function language!"]
    } else {
        set funcbody [string trim [$Win(fnctxt) get 1.0 end]]
        # regsub -all "\n" $funcbody " " funcbody
        regsub -all -- {'} $funcbody {''} funcbody
        regsub -all -- {\\} $funcbody {\\\\} funcbody

        set OK "no"

        if {$saveas_} {

            set sql "CREATE
                   FUNCTION $name
                            ($parameters)
                    RETURNS $returns AS '$funcbody'
                   LANGUAGE '$language'"
            if {[sql_exec noquiet $sql]} {
                set OK "yes"
            }

        } elseif {$nametodrop==$name &&
            $returnstodrop==$returns &&
            $parameterstodrop==$parameters} {
            set sql "CREATE OR REPLACE
                   FUNCTION $name
                            ($parameters)
                    RETURNS $returns AS '$funcbody'
                   LANGUAGE '$language'"
            if {[sql_exec noquiet $sql]} {
                set OK "yes"
            }

        } else {

            set change [tk_messageBox \
                -default no \
                -type yesno \
                -title [intlmsg "Continue ?"] \
                -parent $Win(base) \
                -message [intlmsg "You are to change name, return type or parameters type of function.\nSaving will change function OID."]]

            if {[string match $change "yes"]} {

                sql_exec noquiet "BEGIN TRANSACTION"

                if {$nametodrop != ""} {
                    set sql "DROP
                         FUNCTION $nametodrop
                                  ($parameterstodrop)"
                    if {! [sql_exec noquiet $sql]} {
                        # return
                    }
                }

                set sql "CREATE
                       FUNCTION $name
                                ($parameters)
                        RETURNS $returns AS '$funcbody'
                       LANGUAGE  '$language'"
                if {[sql_exec noquiet $sql]} {
                    sql_exec noquiet "COMMIT TRANSACTION"
                    set OK "yes"
                    set returnstodrop $returns
                    set parameterstodrop $parameters
                    set nametodrop $name
                } else {
                    sql_exec noquiet "ROLLBACK TRANSACTION"
                }
            }
        }

        if {[string match $OK "yes"]} {
            saveComment
            Mainlib::tab_click Functions
        }

        Syntax::highlight $Win(fnctxt) $language
    }

}; # ::Functions::save


#----------------------------------------------------------
# ::Functions::introspect --
#
#   Given a functionname, returns the SQL needed to recreate it
#
# Arguments:
#   functionname_   name of a function to introspect
#   dbh_            an optional database handle
#
# Returns:
#   insql      the CREATE statement to make this function
#----------------------------------------------------------
#
proc ::Functions::introspect {functionname_ {dbh_ ""}} {

    set insql [::Functions::clone $functionname_ $functionname_ $dbh_]

    return $insql

}; # end proc ::Functions::introspect


#----------------------------------------------------------
# ::Functions::clone --
#
#   Like introspect, only changes the functionname
#
# Arguments:
#   srcfunction_    the original function
#   destfunction_   the clone function
#   dbh_            an optional database handle
#
# Returns:
#   insql       the CREATE statement to clone this function
#----------------------------------------------------------
#
proc ::Functions::clone {srcfunction_ destfunction_ {dbh_ ""}} {

    global CurrentDB

    if {[string match "" $dbh_]} {
        set dbh_ $CurrentDB
    }

    set V [::Database::getPgVersion $dbh_]

    set insql ""

    set srcfunction_ [string map {\" {}} $srcfunction_]
    set destfunction_ [string map {\" {}} $destfunction_]

    # hmm, should be 7.3, but do functions have schema in 7.3 ?
    if { $V < 7.4 } {
        set sql "SELECT *
                   FROM pg_proc
                  WHERE
                    (SELECT SUBSTRING('$srcfunction_'
                       FROM 1
                        FOR POSITION('(' IN '$srcfunction_')-1)=proname)
                    AND
                    (SELECT SUBSTRING('$srcfunction_'
                       FROM POSITION('(' IN '$srcfunction_')+1
                            FOR LENGTH('$srcfunction_')-POSITION('(' IN '$srcfunction_')-1)=OIDVECTORTYPES(proargtypes))"
    } else {
        set nsp [lindex [split $srcfunction_ .] 0]
        set srcfunction_ [lindex [split $srcfunction_ .] 1]
        set sql "SELECT p.*
                   FROM pg_proc p,
                        pg_namespace n
                  WHERE
                        p.pronamespace=n.oid
                    AND
                        n.nspname='$nsp'
                    AND
                    (SELECT SUBSTRING('$srcfunction_'
                       FROM 1
                        FOR POSITION('(' IN '$srcfunction_')-1)=proname)
                    AND
                    (SELECT SUBSTRING('$srcfunction_'
                       FROM POSITION('(' IN '$srcfunction_')+1
                            FOR LENGTH('$srcfunction_')-POSITION('(' IN '$srcfunction_')-1)=OIDVECTORTYPES(proargtypes))"
    }

    wpg_select $dbh_ $sql rec {
        set insql "CREATE FUNCTION [::Database::quoteObject [lindex [split $destfunction_ (] 0]] ([::Functions::getParameters $rec(proargtypes) $rec(pronargs) $dbh_])
\tRETURNS\t[::Functions::getPgType $rec(prorettype) $dbh_]"
        set body [string trim [::Database::quoteSQL [::Functions::getBody $rec(proname) $rec(prolang) $rec(prosrc) $dbh_]]]
        if { [string first "\n" $body] == -1 } {
            append insql "\n\tAS '" $body "'"
        } else {
            append insql "\n\tAS '\n" $body "'"
        }
        append insql "\n\tLANGUAGE '[::Functions::getLanguage $rec(prolang) $dbh_]'\n;"
    }

    return $insql

}; # end proc ::Functions::clone


#----------------------------------------------------------
# ::Functions::getPgType --
#
#   Wrapper around ::Database::getPgType with additional
#   functionality to replace "unknown" types with the
#   function-preferred "opaque" definition.
#
# Parameters:
#   oid_    object id to get type name for
#   dbh_    optional database handle
#
# Returns:
#   text name of the provided oid's type
#----------------------------------------------------------
#
proc ::Functions::getPgType {oid_ {dbh_ ""}} {

    set ret [::Database::getPgType $oid_ $dbh_]

    if {[string match $ret "unknown"]} {
        set ret "opaque"
    }

    return $ret

}; # end proc ::Functions::getPgType


#----------------------------------------------------------
# ::Functions::getLanguage --
#
#   Returns name of language for provided oid.
#
# Parameters:
#   oid_    oid of language to retrieve name for
#   dbh_    optional database handle
#
# Returns:
#   text name of the language
#
# Todo:
#   probably move this into Database namespace
#----------------------------------------------------------
#
proc ::Functions::getLanguage {oid_ {dbh_ ""}} {

    global CurrentDB

    if {[string match "" $dbh_]} {
        set dbh_ $CurrentDB
    }

    set lang ""

    set sql "SELECT lanname
               FROM pg_language
              WHERE oid=$oid_"

    wpg_select $dbh_ $sql rec {
        set lang $rec(lanname)
    }

    return $lang

}; # end proc ::Functions::getLanguage


#----------------------------------------------------------
# ::Functions::getBody --
#
#   Convenience to determine C function bodies properly.
#   For other languages, the return value is the provided
#   src_ parameters.
#
# Parameters:
#   func_   name of function
#   lang_   language function is in
#   src_    body of function
#   dbh_    optional database handle
#----------------------------------------------------------
#
proc ::Functions::getBody {func_ lang_ src_ {dbh_ ""}} {

    global CurrentDB

    if {[string match "" $dbh_]} {
        set dbh_ $CurrentDB
    }

    set body ""

    if {$lang_=="C" || $lang_=="c"} {
        set sql "SELECT probin
                   FROM [::Database::qualifySysTable pg_proc]
                  WHERE proname='$func_'"
        wpg_select $dbh_ $sql rec {
            set body $rec(probin)
        }
    } else {
        set body $src_
    }

    return $body

}; # end proc ::Functions::getBody


#----------------------------------------------------------
# ::Functions::getParameters --
#
#   Spins through provided list of type id's and returns the
#   matching type names as a list.
#
# Parameters:
#   argtypes_   list of type id's
#   argcount_   how many type id's were provided
#   dbh_        optional database handle
#
# Returns:
#   comma-separated type names
#----------------------------------------------------------
#
proc ::Functions::getParameters {argtypes_ argcount_ {dbh_}} {

    set params {}

    for {set i 0} {$i<$argcount_} {incr i} {
        lappend params [::Database::getPgType [lindex $argtypes_ $i]]
    }

    set params [join $params ,]

    return $params

}; # end proc ::Functions::getParameters


#----------------------------------------------------------
# ::Functions::saveComment --
#
#   saves the comment on a function
#
# Arguments:
#   delete_     if 1, sets comment string to NULL,
#               thus deleting the comment. default 0
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Functions::saveComment {{delete_ 0}} {

    variable Win
    variable name
    variable parameters
    variable comment

    set sql "COMMENT ON FUNCTION [::Database::quoteObject $name] ($parameters)"

    if {$delete_} {
        append sql " IS NULL"
        $Win(comment) delete 1.0 end
    } else {
        append sql " IS '[string trim [$Win(comment) get 1.0 end]]'"
    }

    sql_exec noquiet $sql

}; # end proc ::Functions::saveComment



#==========================================================
# END NAMESPACE CODE, BEGIN VISUAL TCL
#==========================================================



proc vTclWindow.pgaw:Function {base} {

    global PgAcVar

    if {$base == ""} {
        set base .pgaw:Function
    }

    if {[winfo exists $base]} {
        wm deiconify $base; return
    }

    toplevel $base -class Toplevel
    wm focusmodel $base passive
    wm geometry $base 680x400+98+212
    wm overrideredirect $base 0
    wm resizable $base 1 1
    wm deiconify $base
    wm title $base [intlmsg "Function"]

    # some helpful key bindings
    bind $base <Control-Key-w> [subst {destroy $base}]

    bind $base <Key-F1> "Help::load functions"

    frame $base.fr \
        -borderwidth 10
    pack $base.fr \
        -in $base \
        -expand 1 \
        -fill both
    # make the base setting a little easier
    set ::Functions::Win(base) $base.fr

    Label $::Functions::Win(base).lname \
        -text [intlmsg "Name"]
    set ::Functions::Win(name) [Entry $::Functions::Win(base).ename \
        -textvariable ::Functions::name]

    Label $::Functions::Win(base).lparams \
        -text [intlmsg "Parameters"]
    set ::Functions::Win(parameters) [Entry $::Functions::Win(base).eparams \
        -textvariable ::Functions::parameters]

    Label $::Functions::Win(base).lreturns \
        -text [intlmsg "Returns"]
    set ::Functions::Win(returns) [Entry $::Functions::Win(base).ereturns \
        -textvariable ::Functions::returns]

    Label $::Functions::Win(base).llang \
        -text [intlmsg "Language"]
    set ::Functions::Win(language) [ComboBox $::Functions::Win(base).cblang \
        -textvariable ::Functions::language \
        -values [join [::Database::getLanguagesList]] \
        -editable true]

    set ::Functions::Win(fnctxt) $::Functions::Win(base).text
    text $::Functions::Win(fnctxt) \
        -font $PgAcVar(pref,font_fix) \
        -height 16 \
        -undo 1 \
        -tabs {20 40 60 80 100 120 140 160 180 200} \
        -width 43 \
        -wrap none \
        -yscrollcommand \
            [list $::Functions::Win(base).vscroll set] \
        -xscrollcommand \
            [list $::Functions::Win(base).hscroll set]
    bind $::Functions::Win(fnctxt) <Control-Key-s> {
        Functions::save
    }

    scrollbar $::Functions::Win(base).vscroll \
        -command "$::Functions::Win(fnctxt) yview" \
        -orient vert
    scrollbar $::Functions::Win(base).hscroll \
        -command "$::Functions::Win(fnctxt) xview" \
        -orient horiz

    ButtonBox $::Functions::Win(base).bboxmenu \
        -orient horizontal \
        -homogeneous 1 \
        -spacing 2
    $::Functions::Win(base).bboxmenu add \
        -relief link \
        -image ::icon::fileopen-22 \
        -helptext [intlmsg "Browse"] \
        -borderwidth 1 \
        -command {
            set types {
                {{Compiled Functions}   {.so}}
                {{All Files}            *}
            }
            catch {tk_getOpenFile \
                -parent $::Functions::Win(base) \
                -filetypes $types \
                -title [intlmsg "Browse for Compiled Functions"] \
            } filename
            $::Functions::Win(fnctxt) delete 1.0 end
            $::Functions::Win(fnctxt) insert end $filename
        }
    $::Functions::Win(base).bboxmenu add \
        -relief link \
        -image ::icon::filesave-22 \
        -helptext [intlmsg "Save"] \
        -borderwidth 1 \
        -command {::Functions::save}
    $::Functions::Win(base).bboxmenu add \
        -relief link \
        -image ::icon::filesaveas-22 \
        -helptext [intlmsg "Save As"] \
        -borderwidth 1 \
        -command {::Functions::save 1}
    $::Functions::Win(base).bboxmenu add \
        -relief link \
        -image ::icon::help-22 \
        -helptext [intlmsg "Help"] \
        -borderwidth 1 \
        -command {::Help::load functions}
    $::Functions::Win(base).bboxmenu add \
        -relief link \
        -image ::icon::exit-22 \
        -helptext [intlmsg "Close"] \
        -borderwidth 1 \
        -command {
            catch {Window destroy .pgaw:Function}
        }

    # comment
    TitleFrame $::Functions::Win(base).comment \
        -text [intlmsg "Comments"] \
        -font $PgAcVar(pref,font_bold)
    set titleframe [$::Functions::Win(base).comment getframe]
    set ::Functions::Win(comment) $titleframe.text
    text $titleframe.text \
        -undo 1 \
        -height 3 \
        -width 3
    ButtonBox $titleframe.btnbox \
        -orient vertical \
        -homogeneous 1
    $titleframe.btnbox add \
        -relief link \
        -image ::icon::filesave-22 \
        -helptext [intlmsg "Save"] \
        -borderwidth 1 \
        -command {::Functions::saveComment 0}
    $titleframe.btnbox add \
        -relief link \
        -image ::icon::edittrash-22 \
        -helptext [intlmsg "Delete"] \
        -borderwidth 1 \
        -command {::Functions::saveComment 1}
    grid $titleframe.text \
        -row 0 \
        -column 0 \
        -sticky news
    grid $titleframe.btnbox \
        -row 0 \
        -column 1 \
        -sticky e
    grid columnconfigure $titleframe 0 \
        -weight 10

    grid $::Functions::Win(base).lname \
        -row 0 \
        -column 0 \
        -pady 2 \
        -padx 5 \
        -sticky e
    grid $::Functions::Win(base).ename \
        -row 0 \
        -column 1 \
        -sticky we
    grid $::Functions::Win(base).lparams \
        -row 1 \
        -column 0 \
        -pady 2 \
        -padx 5 \
        -sticky e
    grid $::Functions::Win(base).eparams \
        -row 1 \
        -column 1 \
        -sticky we
    grid $::Functions::Win(base).lreturns \
        -row 0 \
        -column 2 \
        -sticky e
    grid $::Functions::Win(base).ereturns \
        -row 0 \
        -column 3 \
        -sticky we
    grid $::Functions::Win(base).llang \
        -row 1 \
        -column 2 \
        -sticky e
    grid $::Functions::Win(base).cblang \
        -row 1 \
        -column 3 \
        -sticky we

    grid $::Functions::Win(base).bboxmenu \
        -row 0 \
        -column 4 \
        -columnspan 2 \
        -rowspan 2 \
        -sticky e

    grid $::Functions::Win(fnctxt) \
        -row 2 \
        -column 0 \
        -columnspan 5 \
        -sticky news
    grid $::Functions::Win(base).hscroll \
        -row 3 \
        -column 0 \
        -columnspan 5 \
        -sticky new
    grid $::Functions::Win(base).vscroll \
        -row 2 \
        -column 5 \
        -sticky nws

    grid $::Functions::Win(base).comment \
        -row 4 \
        -column 0 \
        -columnspan 5 \
        -sticky sew

    grid columnconfigure $::Functions::Win(base) 1 \
        -weight 5
    grid columnconfigure $::Functions::Win(base) 3 \
        -weight 5
    grid columnconfigure $::Functions::Win(base) 4 \
        -weight 10
    grid rowconfigure $::Functions::Win(base) 2 \
        -weight 10

}


