#==========================================================
# Help --
#
#   provides a hyperlinked help system
#
#==========================================================
#
namespace eval Help {}


#----------------------------------------------------------
# ::Help::findLink
#
#   determines if a tag is a link or not, loads it if it is
#
# Arguments:
#   none
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Help::findLink {} {

    foreach tagname [.pgaw:Help.f.t tag names current] {
        if {$tagname!="link"} {
            ::Help::load $tagname
            return
        }
    }

}

#----------------------------------------------------------
# ::Help::replace_escaped_items
#
#   replaces all escaped items in a string
#
# Arguments:
#   str: the string to modify
#
# Returns:
#   the modified string
#----------------------------------------------------------
#
proc ::Help::replace_escaped_items {str} {
    set index [string first "\\" $str 0]
    while {$index != -1} {
    set str [string replace $str $index $index]
        set index [string first "\\" $str 0]
    }
    return $str
}

# debug purposes
proc ::Help::is {str} {
#    .pgaw:Help.f.t insert end "$str\n"
}

#----------------------------------------------------------
# ::Help::write
#
#   writes a string in the help window with specific tag
#
# Arguments:
#   str: the string to display
#   tag: the format of the string
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Help::write {str {tag normal}} {
    set str [replace_escaped_items $str]
    .pgaw:Help.f.t insert end $str $tag
}

#----------------------------------------------------------
# ::Help::display_tex
#
#   parses and displays a tex line
#
# Arguments:
#   str: string to parse
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Help::display_tex {str} {
    if {[regexp {^[ \t]*[\\]([a-zA-Z]+)[\[]([a-zA-Z0-9]+)[\]][\{]([a-zA-Z]+)[\}][ \t]*(.*)$} \
        $str match command variable body remainder] == 1} {
        is "\\command\[variable\]\{body\}"
        switch $command {
        "documentclass" {}
        }
        write " "
        if {$remainder != ""} {display_tex $remainder}
        return
    }
    if {[regexp {^[ \t]*[\\]newcommand[\{][\\]([a-zA-Z]+)[\}][ \t]*(.*)$} \
        $str match command name remainder] == 1} {
        is "\\newcommand\{\\name\}..."
        switch $command {
        "newcommand"    {}
        }
        return
    }
    if {[regexp {^[ \t]*[\\]([a-zA-Z]+)[\{]([^\}]+)[\}][\{]([-_a-zA-Z0-9]+)[\}][ \t]*(.*)$} \
        $str match command body1 body2 remainder] == 1} {
        is "\\command\{body1\]\{body2\}"
        switch $command {
        "docinsert"    {write $body1 "link $body2"}
        "htmlref"    {write $body1 "link $body2"}
        }
        write " "
        if {$remainder != ""} {display_tex $remainder}
        return
    }
    if {[regexp {^[ \t]*[\\]([a-zA-Z]+)[\{]([^\}]*)[\}][ \t]*(.*)$} \
    $str match command body remainder] == 1} {
        is "\\command\{body\}"
        switch $command {
        "documentclass"    {}
        "usepackage"    {}
        "input"        {}
        "label"        {}
        "begin"        {}
        "end"        {
                    switch $body {
                    "itemize"    {write "\n\t"}
                    "description"    {write "\n\t"}
                    }
                }
        "title"        {write "\n"; write $body maintitle; write "\n\n\t"}
        "section"    {write "\n"; write $body title; write "\n\n\t"}
        "subsection"    {write "\n"; write $body subtitle; write "\n\n\t"}
        "emph"        {write $body italic}
        "textbf"    {write $body bold}
        "textit"    {write $body italic}
        "texttt"    {write $body code}
        "item"        {write "\n\t\t$body " italic}
        }
        write " "
        if {$remainder != ""} {display_tex $remainder}
        return
    }
    if {[regexp {^[ \t]*[\\]([a-zA-Z]+)[\[]([^\]]+)[\]][ \t]*(.*)$} \
    $str match command body remainder] == 1} {
        is "\\command\[body\]"
        switch $command {
        "item"    {write "\n\t\t$body " italic}
        }
        write " "
        if {$remainder != ""} {display_tex $remainder}
        return
    }
    if {[regexp {^[ \t]*[\\]([a-zA-Z]+)[ \t]*(.*)$} \
    $str match command remainder] == 1} {
        is "\\command ..."
        switch $command {
        "maketitle"        {}
        "tableofcontents"    {}
        "newpage"        {}
        "topics"        {write "Topics available:\n"}
        "topic"            {write "\n\t\t* " italic}
        "LaTeX"            {write LaTeX}
        "item"            {write "\n\t\t* " italic}
        }
        write " "
        if {$remainder != ""} {display_tex $remainder}
        return
    }
    if {[regexp {^[ \t]*[%]} \
    $str match] == 1} {
        is "comment"
        return
    }
    if {[regexp {^[ \t]*$} \
    $str match] == 1} {
        is "empty line"
        write "\n\n\t"
        return
    }
    if {[regexp {^[ \t]*([[:alnum:][:punct:][:graph:]]+)[ \t]*(.*)$} \
    $str match word remainder] == 1} {
        is "plain text"
        write $word
        write " "
        if {$remainder != ""} {display_tex $remainder}
        return
    }
    
    write $str
    write " "
    return
}


#----------------------------------------------------------
# ::Help::display_html --
#
#   a callback for ::htmlparse::parse that handles the
#   parsed HTML help data
#
# Arguments:
#   tag_    name of the tag currently being processed
#   close_  empty if start of tag, char "/" if end of tag
#   param_  list of parameters to the tag
#   txt_    data in between tags
#
# Returns:
#   none
#
# Modifes:
#   the help system main text widget
#----------------------------------------------------------
#
proc ::Help::display_html {tag_ close_ param_ txt_} {
    variable linkanchors
    variable lastclass
    variable lastclassdiv
    variable lastlinktxt
    variable lastlinkurl

    set txt_ [::htmlparse::mapEscapes $txt_]
    set param_ [string trim $param_]
    if {[string match "CLASS=*" $param_]} {
        if {$close_ == ""} {
            set lastclass [lindex [split $param_ {\".}] 1]
        } else {
            set lastclass ""
        }
    }

    switch $tag_ {
        TITLE {
            set linkanchors 0
            write "\n"
            write $txt_ title
        }
        DIV {
            if {$close_ == ""} {
                set lastclassdiv $lastclass
            } else {
                set lastclassdiv ""
            }
            if {$lastclass == "TOC"} {
                if {$close_ == ""} {
                    set linkanchors 1
                    write "\n\n"
                    write "Table of Contents" subtitle
                    write "\n\n"
                } else {
                    set linkanchors 0
                }
            } elseif {$lastclass == "NAVFOOTER"} {
                set linkanchors 0
            } elseif {$lastclass == "REFSYNOPSISDIV"} {
                set linkanchors 1
            } elseif {$lastclass == "SECT1"} {
                set linkanchors 1
            }
        }
        A {
            if {$linkanchors} {
                if {$close_ == ""} {
                    if {[string match "*HREF*" $param_]} {
                        set lastlinkurl [lindex [split $param_ {\".}] 1]
                        set lastlinktxt $txt_
                    } else {
                        write $txt_
                    }
                } else {
                    if {[string trim $lastlinkurl] != ""} {
                        write $lastlinktxt "link $lastlinkurl"
                        write [string map {\n { } {  } {}} $txt_]
                        set lastlinktxt ""
                        set lastlinkurl ""
                    } else {
                        write [string map {\n { } {  } {}} $txt_]
                    }
                }
            }
        }
        ACRONYM {
            if {$close_ == ""} {
                write [string map {\n { } {  } {}} $txt_]
            } else {
                write [string map {\n { } {  } {}} $txt_]
            }
        }
        SPAN {
            if {$close_ == ""} {
                write [string map {\n { } {  } {}} $txt_]
            } else {
                write [string map {\n { } {  } {}} $txt_]
            }
        }
        DL {
            if {$close_ != ""} {
                write "\n"
            }
        }
        BR {
            if {$close_ == ""} {
                write "\n"
            }
        }
        DD {
            if {$close_ != ""} {
                write "\n"
            }
        }
        DT {
            if {$close_ == ""} {
                if {$lastclassdiv != "VARIABLELIST"} {
                    write "\n"
                }
            }
        }
        H1 {
            if {$close_ == ""} {
                write "\n"
            } else {
                write "\n\n"
            }
        }
        H3 {
            if {$close_ != ""} {
                write "\n"
            }
        }
        H2 {
            if {$close_ == ""} {
                write "\n\n"
                write $txt_ subtitle
                write "\n\n"
            } elseif {$lastclassdiv == "REFNAMEDIV"} {
                write $txt_
            }
        }
        I {
            if {$close_ == ""} {
                if {[string trim $lastlinkurl] != ""} {
                    write $txt_ "link $lastlinkurl"
                    set lastlinkurl ""
                } else {
                    write $txt_ italic
                }
            } elseif {[string trim $txt_] != ""} {
                if {[string trim $lastlinkurl] != ""} {
                    write $txt_ "link $lastlinkurl"
                    set lastlinkurl ""
                } else {
                    write [string map {\n { } {  } {}} $txt_]
                }
            }
        }
        PRE {
            if {$close_ == ""} {
                if {$lastclass == "PROGRAMLISTING"} {
                    write "\n\t$txt_\n" code
                } else {
                    write $txt_ code
                }
            } else {
                if {$lastclass == "PROGRAMLISTING"} {
                    write "\n[string map {\n { } {  } {}} $txt_]"
                } else {
                    write "\n\t$txt_"
                }
            }
        }
        TT {
            if {$close_ == ""} {
                if {$lastclassdiv == "TABLE"} {
                    write "[string map {\n { } {  } {}} $txt_]\t\t\t\t\t"
                } elseif {$lastclassdiv == "REFSECT2"} {
                    write [string map {\n { } {  } {}} $txt_] bold
                } elseif {$lastclass == "LITERAL"} {
                    write [string map {\n { } {  } {}} $txt_]
                } else {
                    if {[string trim $txt_] != ""} {
                        write $txt_ code
                    }
                }
            } else {
                if {$lastclassdiv == "REFSYNOPSISDIV"} {
                    write $txt_ code
                } elseif {$lastclassdiv == "VARIABLELIST"} {
                    write [string map {\n { } {  } {}} $txt_]
                } else {
                    write [string map {\n { } {  } {}} $txt_]
                }
            }
        }
        P {
            if {$close_ == ""} {
                if {$lastclassdiv == "VARIABLELIST"} {
                    write "\n\t[string map {\n { } {  } {}} $txt_]"
                } else {
                    write [string map {\n { } {  } {}} $txt_]
                }
            } else {
                write "\n\n"
            }
        }
        TH {
            if {$close_ == ""} {
                if {$lastclassdiv == "TABLE"} {
                    write "[string map {\n { } {  } {}} $txt_]\t\t" bold
                } else {
                    write "[string map {\n { } {  } {}} $txt_]\n" maintitle
                }
            } else {
                if {$lastclassdiv == "TABLE"} {
                    write "\t\t"
                }
            }
        }
        TR {
            if {$close_ == "" && $lastclassdiv == "TABLE"} {
                write "\n"
            }
        }
        TD {
            if {$close_ == "" && $lastclassdiv == "TABLE"} {
                if {[string trim $txt_] != ""} {
                    write "|[string map {\n { } {  } {}} $txt_]|\t\t\t\t"
                }
            }
        }
        CODE {
            if {$close_ == ""} {
                write [string map {\n { } {  } {}} $txt_] code
            } else {
                write [string map {\n { } {  } {}} $txt_]
            }
        }
        SAMP {
            if {$close_ == ""} {
                write [string map {\n { } {  } {}} $txt_] code
            } else {
                write [string map {\n { } {  } {}} $txt_]
            }
        }
        KBD {
            if {$close_ == ""} {
                write [string map {\n { } {  } {}} $txt_] code
            } else {
                write $txt_
            }
        }
        LI {
            if {$close_ == ""} {
                write "*" bold
            }
        }
        B {
            if {![string match "Table of Contents" $txt_]} {
                if {$close_ == ""} {
                    write [string map {\n { } {  } {}} $txt_] bold
                } else {
                    write $txt_
                }
            }
        }
        default {
            # ignored
            # HR, TABLE, BODY, HTML
            #puts "$tag_ : $txt_"
            #write [string map {\n { } {  } {}} $txt_]
        }
    }

}; # end proc ::Help::display_html


#----------------------------------------------------------
# ::Help::load
#
#   loads the help info for a link
#
# Arguments:
#
#   topic_  the help topic to load
#   args_   name of the previous help topic, if any
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Help::load {topic_ {args_ ""}} {

    global PgAcVar
    variable linkanchors
    variable lastclass
    variable lastclassdiv
    variable lastlinkurl
    variable lastlinktxt

    if {![winfo exists .pgaw:Help]} {
        Window show .pgaw:Help
        tkwait visibility .pgaw:Help
    }
    wm deiconify .pgaw:Help

    .pgaw:Help.fb.tex configure -text ""

    if {![info exists PgAcVar(help,history)]} {
        set PgAcVar(help,history) {}
    }

    if {[llength $args_]==1} {
        set PgAcVar(help,current_topic) [lindex $args_ 0]
        set PgAcVar(help,history) [lrange $PgAcVar(help,history) 0 [lindex $args_ 0]]
    } else {
        lappend PgAcVar(help,history) $topic_
        set PgAcVar(help,current_topic) [expr {[llength $PgAcVar(help,history)]-1}]
    }

    # Limit the history length to 100 topics
    if {[llength $PgAcVar(help,history)]>100} {
        set PgAcVar(help,history) [lrange $PgAcVar(help,history) 1 end]
    }

    .pgaw:Help.f.t configure -state normal
    .pgaw:Help.f.t delete 1.0 end
    .pgaw:Help.f.t tag configure normal -font $PgAcVar(pref,font_normal)
    .pgaw:Help.f.t tag configure bold -font $PgAcVar(pref,font_bold)
    .pgaw:Help.f.t tag configure italic -font $PgAcVar(pref,font_italic)
    .pgaw:Help.f.t tag configure large -font {Helvetica -14 bold}
    .pgaw:Help.f.t tag configure title -font $PgAcVar(pref,font_bold) -justify center
    .pgaw:Help.f.t tag configure maintitle -font {Helvetica -16 bold} -justify center
    .pgaw:Help.f.t tag configure subtitle -font $PgAcVar(pref,font_bold)
    .pgaw:Help.f.t tag configure link -font {Helvetica -14 underline} -foreground #000080
    .pgaw:Help.f.t tag configure code -font $PgAcVar(pref,font_fix)
    .pgaw:Help.f.t tag configure warning -font $PgAcVar(pref,font_bold) -foreground #800000
    .pgaw:Help.f.t tag bind link <Button-1> {Help::findLink}
    set errmsg {}
    .pgaw:Help.f.t configure -tabs {30 60 90 120 150 180 210 240 270 300 330 360 390}
    catch {
        # TODO add option to view local postgres docs,
        # ie. those obtained via apt-get install postgresql-docs
        if {[file exists [file join $PgAcVar(PGACCESS_DOC) pg $topic_.html]]} {
            set fd [open [file join $PgAcVar(PGACCESS_DOC) pg $topic_.html] r]
            set str [read $fd]
            close $fd
            set linkanchors 0
            set lastclass ""
            set lastclassdiv ""
            set lastlinkurl ""
            set lastlinktxt ""
            ::htmlparse::parse -cmd ::Help::display_html $str
        } elseif {! [file exists [file join $PgAcVar(PGACCESS_DOC) tex $topic_.tex]]} {
            source [file join $PgAcVar(PGACCESS_DOC) help $topic_.hlp]
        } else {
            set fd [open [file join $PgAcVar(PGACCESS_DOC) tex $topic_.tex] r]
            set str [read $fd]
            close $fd
            #.pgaw:Help.fb.tex configure -text "tex"
            set lines [split $str \n]
            foreach l $lines {
                display_tex $l
            }
        }
    } errmsg
    if {$errmsg!=""} {
        .pgaw:Help.f.t insert end "Error loading help file [file join $PgAcVar(PGACCESS_DOC) help $topic_.hlp]\n\n$errmsg\n\n\n\nPlease understand that PgAccess does not contain all of the documentation for PostgreSQL.  Complete documentation can be downloaded from www.postgresql.org" bold
    }
    .pgaw:Help.f.t configure -state disabled
    focus .pgaw:Help.f.sb
}


#----------------------------------------------------------
# ::Help::back --
#
#   moves backwards in the help topics list
#
# Arguments:
#   none
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Help::back {} {
global PgAcVar

    if {![info exists PgAcVar(help,history)]} {return}
    if {[llength $PgAcVar(help,history)]==0} {return}
    set i $PgAcVar(help,current_topic)
    if {$i<1} {return}
    incr i -1

    ::Help::load [lindex $PgAcVar(help,history) $i] $i

}



#==========================================================
# END HELP NAMESPACE
# BEGIN VISUAL TCL CODE
#==========================================================



proc vTclWindow.pgaw:Help {base} {
global PgAcVar

    if {$base == ""} {
        set base .pgaw:Help
    }

    if {[winfo exists $base]} {
        wm deiconify $base; return
    }

    toplevel $base -class Toplevel
    wm focusmodel $base passive
    set sw [winfo screenwidth .]
    set sh [winfo screenheight .]
    set x [expr {($sw - 640)/2}]
    set y [expr {($sh - 480)/2}]

    wm geometry $base 640x480+$x+$y
    wm maxsize $base 1280 1024
    wm minsize $base 1 1
    wm overrideredirect $base 0
    wm resizable $base 1 1
    wm deiconify $base
    wm title $base [intlmsg "Help"]
    bind $base <Key-Escape> "Window destroy .pgaw:Help"

    frame $base.fb \
        -borderwidth 2 -height 75 -relief flat -width 125
    button $base.fb.bback \
        -command Help::back -padx 9 -pady 3 -text [intlmsg Back]
    button $base.fb.bi \
        -command {Help::load index} -padx 9 -pady 3 -text [intlmsg Index]
    button $base.fb.bp \
        -command {Help::load postgresql} -padx 9 -pady 3 -text PostgreSQL
    button $base.fb.btnclose \
        -command {Window destroy .pgaw:Help} -padx 9 -pady 3 -text [intlmsg Close]
    frame $base.f \
        -borderwidth 2 -height 75 -relief flat -width 125
    text $base.f.t \
        -borderwidth 1 -cursor {} -font $PgAcVar(pref,font_normal) -height 2 \
        -highlightthickness 0 -state disabled \
        -tabs {30 60 90 120 150 180 210 240 270 300 330 360 390} -width 8 \
        -wrap word -yscrollcommand {.pgaw:Help.f.sb set} \
        -background white
    scrollbar $base.f.sb \
        -borderwidth 1 -command {.pgaw:Help.f.t yview} -highlightthickness 0 \
        -orient vert

    pack $base.fb \
        -in .pgaw:Help -anchor center -expand 0 -fill x -side top
    pack $base.fb.bback \
        -in .pgaw:Help.fb -anchor center -expand 0 -fill none -side left
    pack $base.fb.bi \
        -in .pgaw:Help.fb -anchor center -expand 0 -fill none -side left
    pack $base.fb.bp \
        -in .pgaw:Help.fb -anchor center -expand 0 -fill none -side left
#####
# just for debug purposes
    label $base.fb.tex
    pack $base.fb.tex \
        -in .pgaw:Help.fb -anchor center -expand 0 -fill none -side left
#####
    pack $base.fb.btnclose \
        -in .pgaw:Help.fb -anchor center -expand 0 -fill none -side right
    pack $base.f \
        -in .pgaw:Help -anchor center -expand 1 -fill both -side top
    pack $base.f.t \
        -in .pgaw:Help.f -anchor center -expand 1 -fill both -side left
    pack $base.f.sb \
        -in .pgaw:Help.f -anchor center -expand 0 -fill y -side right

}

