#==========================================================
# Sequences --
#
#   provides rudimentary access to database sequences
#
#==========================================================
#
namespace eval Sequences {
    variable Win
    variable name
    variable increment
    variable start
    variable minval
    variable maxval
    variable comment
    variable cycle
    variable cache
}


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Sequences::new {} {

    variable Win
    variable name ""
    variable increment 1
    variable start 1
    variable minval 1
    variable maxval 9223372036854775807
    variable comment ""
    variable cycle "f"
    variable cache 1

    Window show .pgaw:Sequence

    $Win(comment) delete 1.0 end

    focus $Win(name)

}; # end proc ::Sequences::new


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Sequences::design {seqname_} {
    ::Sequences::open $seqname_
}; # end proc ::Sequences::design


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Sequences::open {seqname_} {

    global CurrentDB

    variable Win
    variable name
    variable increment
    variable start
    variable minval
    variable maxval
    variable comment
    variable cycle
    variable cache

    Window show .pgaw:Sequence

    set flag 1

    # getting the comment requires an oid which means we
    # need to worry about PG system tables, thus, versioning
    set V [::Database::getPgVersion $CurrentDB]

    if {$V<7.3} {
        set sql "
            SELECT s.*,
                   obj_description(c.oid)
                AS sdesc
              FROM [::Database::quoteObject $seqname_] s,
                   [::Database::qualifySysTable pg_class $CurrentDB] c
             WHERE (relkind ='S')
               AND s.sequence_name=c.relname"
    } else {
        set sql "
            SELECT s.*,
                   obj_description(c.oid)
                AS sdesc
              FROM [::Database::quoteObject $seqname_] s,
                   pg_catalog.pg_class c,
                   pg_catalog.pg_namespace n
             WHERE (relkind ='S')
               AND n.oid=c.relnamespace
               AND s.sequence_name=c.relname"
    }

    wpg_select $CurrentDB $sql rec {
        set flag 0
        set name $seqname_
        set increment $rec(increment_by)
        set start $rec(last_value)
        set minval $rec(min_value)
        set maxval $rec(max_value)
        set cache $rec(cache_value)
        set cycle $rec(is_cycled)
        set comment $rec(sdesc)
        $Win(comment) insert end $comment
        $Win(lstart) configure \
            -text [intlmsg "Last value"]
    }

    if {$flag} {
        showError [format [intlmsg "Sequence '%s' not found!"] $seqname_]
    }

}; # end proc ::Sequences::open


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Sequences::save {} {

    variable Win
    variable name
    variable increment
    variable start
    variable minval
    variable maxval
    variable cycle
    variable cache

    if {$name==""} {
        showError [intlmsg "You should supply a name for this sequence"]
    } else {
        set sql "CREATE SEQUENCE [::Database::quoteObject $name]"
        if {$increment!=""} {
            append sql " INCREMENT $increment"
        }
        if {$start!=""} {
            append sql " START $start"
        }
        if {$minval!=""} {
            append sql " MINVALUE $minval"
        }
        if {$maxval!=""} {
            append sql " MAXVALUE $maxval"
        }
        if {$cache!=""} {
            append sql " CACHE $cache"
        }
        if {$cycle=="t"} {
            append sql " CYCLE"
        }

        if {[sql_exec noquiet $sql]} {
            saveComment
            ::Mainlib::tab_click Sequences
            tk_messageBox \
                -title [intlmsg Information] \
                -parent $Win(base) \
                -message [intlmsg "Sequence created!"]
        }

    }

}; # end proc ::Sequences::save


#----------------------------------------------------------
# ::Sequences::saveComment --
#
#   saves the comment on a sequence
#
# Arguments:
#   delete_     if 1, sets comment string to NULL,
#               thus deleting the comment. default 0
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Sequences::saveComment {{delete_ 0}} {

    variable Win
    variable name
    variable comment

    set sql "COMMENT ON SEQUENCE [::Database::quoteObject $name]"

    if {$delete_} {
        append sql " IS NULL"
        $Win(comment) delete 1.0 end
    } else {
        append sql " IS '[string trim [$Win(comment) get 1.0 end]]'"
    }

    sql_exec noquiet $sql

}; # end proc ::Sequences::saveComment


#----------------------------------------------------------
# ::Sequences::introspect --
#
#   Given a sequencename, returns the SQL needed to recreate it
#
# Arguments:
#   sequencename_   name of a sequence to introspect
#   dbh_            an optional database handle
#
# Returns:
#   insql      the CREATE statement to make this sequence
#----------------------------------------------------------
#
proc ::Sequences::introspect {sequencename_ {dbh_ ""}} {

    set insql [::Sequences::clone $sequencename_ $sequencename_ $dbh_]

    return $insql

}; # end proc ::Sequences::introspect


#----------------------------------------------------------
# ::Sequences::clone --
#
#   Like introspect, only changes the sequencename
#
# Arguments:
#   srcsequence_    the original sequence
#   destsequence_   the clone sequence
#   dbh_            an optional database handle
#
# Returns:
#   insql       the CREATE statement to clone this sequence
#----------------------------------------------------------
#
proc ::Sequences::clone {srcsequence_ destsequence_ {dbh_ ""}} {

    global CurrentDB

    if {[string match "" $dbh_]} {
        set dbh_ $CurrentDB
    }

    set insql ""

    set sql "SELECT *
               FROM [::Database::quoteObject $srcsequence_]"

    wpg_select $dbh_ $sql rec {
        set insql "CREATE SEQUENCE [::Database::quoteObject $destsequence_]
\tINCREMENT\t$rec(increment_by)
\tMINVALUE\t$rec(min_value)
\tMAXVALUE\t$rec(max_value)
\tSTART\t\t$rec(last_value)
\tCACHE\t\t$rec(cache_value)"
        if {$rec(is_cycled)=="t"} {
            append insql "\n\tCYCLE"
        }
        append insql "\n;"
    }

    return $insql

}; # end proc ::Sequences::clone



#==========================================================
# END NAMESPACE SEQUENCES
# BEGIN VISUAL TCL CODE
#==========================================================



proc vTclWindow.pgaw:Sequence {base} {

    global PgAcVar

    if {$base == ""} {
        set base .pgaw:Sequence
    }

    if {[winfo exists $base]} {
        wm deiconify $base
        return
    }

    toplevel $base -class Toplevel
    wm focusmodel $base passive
    wm geometry $base 320x260+120+210
    wm maxsize $base 1009 738
    wm minsize $base 1 1
    wm overrideredirect $base 0
    wm resizable $base 1 1
    wm deiconify $base
    wm title $base [intlmsg "Sequence"]

    # some helpful key bindings
    bind $base <Control-Key-w> [subst {destroy $base}]

    bind $base <Key-F1> "Help::load sequences"

    frame $base.f1 \
        -borderwidth 10
    pack $base.f1 \
        -in $base \
        -expand 1 \
        -fill both
    # make the base setting a little easier
    set ::Sequences::Win(base) $base.f1

    Label $base.f1.l1 \
        -text [intlmsg "Sequence name"]
    set ::Sequences::Win(name) [Entry $base.f1.e1 \
        -textvariable ::Sequences::name]
    bind $base.f1.e1 <Key-KP_Enter> {
        focus .pgaw:Sequence.f1.e2
    }
    bind $base.f1.e1 <Key-Return> {
        focus .pgaw:Sequence.f1.e2
    }

    Label $base.f1.l2 \
        -text [intlmsg "Increment"]
    set ::Sequences::Win(increment) [Entry $base.f1.e2 \
        -textvariable ::Sequences::increment]
    bind $base.f1.e2 <Key-Return> {
        focus .pgaw:Sequence.f1.e3
    }

    set ::Sequences::Win(lstart) [Label $base.f1.l3 \
        -text [intlmsg "Start value"]]
    set ::Sequences::Win(start) [Entry $base.f1.e3 \
        -textvariable ::Sequences::start]
    bind $base.f1.e3 <Key-Return> {
        focus .pgaw:Sequence.f1.e4
    }

    Label $base.f1.l4 \
        -text [intlmsg "Minvalue"]
    set ::Sequences::Win(minval) [Entry $base.f1.e4 \
        -textvariable ::Sequences::minval]
    bind $base.f1.e4 <Key-Return> {
        focus .pgaw:Sequence.f1.e5
    }

    Label $base.f1.l5 \
        -text [intlmsg "Maxvalue"]
    set ::Sequences::Win(maxval) [Entry $base.f1.e5 \
        -textvariable ::Sequences::maxval]
    bind $base.f1.e5 <Key-Return> {
        focus .pgaw:Sequence.f1.e6
    }

    Label $base.f1.l6 \
        -text [intlmsg "Cache"]
    set ::Sequences::Win(cache) [Entry $base.f1.e6 \
        -textvariable ::Sequences::cache]

    set ::Sequences::Win(cycle) $base.f1.cbcycle
    checkbutton $base.f1.cbcycle \
        -text [intlmsg "Cycle"] \
        -onvalue "t" \
        -offvalue "f" \
        -variable ::Sequences::cycle

    ButtonBox $::Sequences::Win(base).bboxmenu \
        -orient horizontal \
        -homogeneous 1 \
        -spacing 2
    $::Sequences::Win(base).bboxmenu add \
        -relief link \
        -image ::icon::filesave-22 \
        -borderwidth 1 \
        -command {Sequences::save} \
        -helptext [intlmsg "Define sequence"]
    $::Sequences::Win(base).bboxmenu add \
        -relief link \
        -image ::icon::help-22 \
        -borderwidth 1 \
        -command {::Help::load sequences} \
        -helptext [intlmsg "Help"]
    $::Sequences::Win(base).bboxmenu add \
        -relief link \
        -image ::icon::exit-22 \
        -borderwidth 1 \
        -command {Window destroy .pgaw:Sequence} \
        -helptext [intlmsg "Close"]

    # comment
    TitleFrame $::Sequences::Win(base).comment \
        -text [intlmsg "Comments"] \
        -font $PgAcVar(pref,font_bold)
    set titleframe [$::Sequences::Win(base).comment getframe]
    set ::Sequences::Win(comment) $titleframe.text
    text $titleframe.text \
        -height 3 \
        -width 3
    ButtonBox $titleframe.btnbox \
        -orient vertical \
        -homogeneous 1
    $titleframe.btnbox add \
        -relief link \
        -image ::icon::filesave-22 \
        -helptext [intlmsg "Save"] \
        -borderwidth 1 \
        -command {::Sequences::saveComment 0}
    $titleframe.btnbox add \
        -relief link \
        -image ::icon::edittrash-22 \
        -helptext [intlmsg "Delete"] \
        -borderwidth 1 \
        -command {::Sequences::saveComment 1}
    grid $titleframe.text \
        -row 0 \
        -column 0 \
        -sticky news
    grid $titleframe.btnbox \
        -row 0 \
        -column 1 \
        -sticky e
    grid columnconfigure $titleframe 0 \
        -weight 10

    grid $base.f1.l1 \
        -column 0 \
        -row 0 \
        -sticky w
    grid $base.f1.e1 \
        -column 1 \
        -row 0
    grid $base.f1.l2 \
        -column 0 \
        -row 2 \
        -sticky w
    grid $base.f1.e2 \
        -column 1 \
        -row 2
    grid $base.f1.l3 \
        -column 0 \
        -row 4 \
        -sticky w
    grid $base.f1.e3 \
        -column 1 \
        -row 4
    grid $base.f1.l4 \
        -column 0 \
        -row 6 \
        -sticky w
    grid $base.f1.e4 \
        -column 1 \
        -row 6
    grid $base.f1.l5 \
        -column 0 \
        -row 7 \
        -sticky w
    grid $base.f1.e5 \
        -column 1 \
        -row 7
    grid $base.f1.l6 \
        -column 0 \
        -row 8 \
        -sticky w
    grid $base.f1.e6 \
        -column 1 \
        -row 8
    grid $base.f1.cbcycle \
        -column 1 \
        -row 9 \
        -sticky w

    grid $::Sequences::Win(base).bboxmenu \
        -column 2 \
        -row 0 \
        -sticky e

    grid $::Sequences::Win(base).comment \
        -column 0 \
        -row 10 \
        -columnspan 3 \
        -sticky news

}

